//*******************************************************
// VRML++ Preprocessor
// Copyright (C) 1996 Stephan Diehl
// 
// extends the VRML 2.0 Parser of Bell and Woods,
// (see Copyright message below)
// 
// extensions are enclosed in #ifdef STED ... #endif
//            or comments /*<< STED */ ... /* STED >>*/
//
//*******************************************************
// Major Extensions 
//   * building a parse tree while parsing
//   * checking whether USED nodes are DEFined
//   * adding productions for VRML++ features
//   * improved type checking 
//   * conversion of VRML++ features into VRML 2.0
//   * unparsing of parse tree
//*******************************************************
//

//*******************************************************
// VRML 2.0 Parser
// Copyright (C) 1996 Silicon Graphics, Inc.
//
// Author(s)    : Gavin Bell
//                Daniel Woods (first port, minor fixes)
//*******************************************************

%{

// only to debug type checking #define TCHECKDEBUG  1

#define myfree(x) free(x) 

int yydebug=1;

//
// Parser for VRML 2.0 files.
// This is a minimal parser that does NOT generate an in-memory scene graph.
// 

// The original parser was developed on a Windows 95 PC with
// Borland's C++ 5.0 development tools.  This was then ported
// to a Windows 95 PC with Microsoft's MSDEV C++ 4.0 development
// tools.  The port introduced the ifdef's for
//    USING_BORLAND_CPP_5          : since this provides a "std namespace",
//    TWO_ARGUMENTS_FOR_STL_STACK  : STL is a moving target.  The stack template
//                                     class takes either one or two arguments.

#include <iostream.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <list.h>
#include <stack.h>
#include <string.h>

#include "vrmlnode.h"

#ifdef USE_STD_NAMESPACE
using namespace std;
#endif

// Currently-being-define proto.  Prototypes may be nested, so a stack
// is needed:

#ifdef TWO_ARGUMENTS_FOR_STL_STACK  
stack< VrmlNodeType*, list<VrmlNodeType*> > currentProtoStack;
#else
stack< list<VrmlNodeType*> > currentProtoStack;
#endif  /* TWO_ARGUMENTS_FOR_STL_STACK */

// This is used to keep track of which field in which type of node is being
// parsed.  Field are nested (nodes are contained inside MFNode/SFNode fields)
// so a stack of these is needed:
typedef struct {
	const VrmlNodeType *nodeType;
        const char *fieldName;
} FieldRec;
#ifdef TWO_ARGUMENTS_FOR_STL_STACK  
stack< FieldRec*, list<FieldRec*> > currentField;
#else
stack< list<FieldRec*> > currentField;
#endif  /* TWO_ARGUMENTS_FOR_STL_STACK */

// This is used when the parser knows what kind of token it expects
// to get next-- used when parsing field values (whose types are declared
// and read by the parser) and at certain other places:
extern int expectToken;

// Current line number (set by lexer)
extern int currentLineNumber;

extern ParseTreeNode* parsetree;

// Some helper routines defined below:
static void beginProto(const char *);
static void beginClass(const char *);
static void endProto();
static void endClass();
int addField(const char *type, const char *name, ParseTreeNode*);
int addEventIn(const char *type, const char *name);
int addEventOut(const char *type, const char *name);
int addExposedField(const char *type, const char *name, ParseTreeNode* );
int add(void (VrmlNodeType::*)(const char *, int, const char*),
        const char *, const char *);
int add(void (VrmlNodeType::*)(const char *, int, const char*, ParseTreeNode*), 
        const char *, const char *, ParseTreeNode*);
int fieldType(const char *type);
void enterNode(const char *);
void exitNode();
void inScript();
void enterField(const char *);
void exitField();
void expect(int type);
void addSuperclass(const char *);

int hasFieldorEvent(const char *);

ParseTreeNode* tmppt;

/* STED type checking */

typedef struct { char* type;
         char* name;
       } DefNode;

list<DefNode*> deflist;

DefNode* finddef(const char*);
int subtype(const char*, const char*);
int check_subtype(const char*, const char*);
VrmlNodeType::NameTypeRec* getFieldorEvent(const char*);
VrmlNodeType::NameTypeRec* getFieldorEvent(VrmlNodeType*,const char*);

extern "C" void yyerror(const char *msg);

%}


%union {
	char *string;

    /* Other types that will be needed by a true VRML implementation
     * (but are not defined by this parser due to the complexity):
     * Node *node;
     * list<Node *> *nodeList;
     */

     ParseTreeNode* pt;
};

%type <string> IDENTIFIER

%type <pt> vrmlscene declarations declaration nodeDeclaration 
%type <pt> protoDeclaration proto externproto interfaceDeclarations
%type <pt> interfaceDeclaration externInterfaceDeclarations 
%type <pt> externInterfaceDeclaration routeDeclaration routeEvent
%type <pt> node nodeGuts nodeGut fieldValue mfnodeValue nodes
%type <pt> classDeclaration class externclass extends superclasses

%type <pt> SFBOOL SFCOLOR SFFLOAT SFIMAGE SFINT32 SFNODE SFROTATION
%type <pt> SFSTRING SFTIME SFVEC2F SFVEC3F
%type <pt> MFCOLOR MFFLOAT MFINT32 MFROTATION MFSTRING MFVEC2F MFVEC3F
%type <pt> MFNODE

/*
 * And types that will be needed by a true VRML implementation:
 * %type <nodeList> vrmlscene declarations
 */

%token CLASS EXTERNCLASS EXTENDS
%token IDENTIFIER DEF USE PROTO EXTERNPROTO TO IS ROUTE SFN_NULL
%token EVENTIN EVENTOUT FIELD EXPOSEDFIELD UNROUTE
%token SFBOOL SFCOLOR SFFLOAT SFIMAGE SFINT32 SFNODE SFROTATION
%token SFSTRING SFTIME SFVEC2F SFVEC3F
%token MFCOLOR MFFLOAT MFINT32 MFROTATION MFSTRING MFVEC2F MFVEC3F
%token MFNODE

%%


vrmlscene:	declarations { parsetree=$1; }
	;

declarations:
		/* Empty is OK */         { $$=pt_node(PT_DECS,NULL); } 
	|	declarations declaration  { $1->addChild($2); 
                                            $$=$1; }
    ;

declaration:
		nodeDeclaration     {$$=$1;}
	|	protoDeclaration    {$$=$1;}
	|	routeDeclaration    {$$=$1;}
/*<< STED */
        |       classDeclaration    {$$=$1;}
/* STED >>*/
    ;

nodeDeclaration: node
	|	DEF IDENTIFIER node		{ DefNode* dp=new DefNode;
                                                  $$=pt_node(PT_DEF,$2,$3);
                                                  myfree($2);
                                                  $$->classname=$3->classname;
                                                  dp->type=$3->classname;
                                                  dp->name=strdup($2);
                                                  deflist.push_front(dp); }
        |	USE IDENTIFIER			
                 { DefNode* dp=finddef($2);
                   $$=pt_node(PT_USE,$2);
                   if (dp)
                     $$->classname=dp->type;
                   else
                    { yyerror("No previous DEF "); 
                      cerr << $2 <<"\n";
                      $$->classname="UNKNOWN";
                    }
                   myfree($2); }
        ;

protoDeclaration:       proto         {$$=$1;}
	       |	externproto   {$$=$1;}
    ;

proto: PROTO IDENTIFIER                   { beginProto($2); }	  
      	'[' interfaceDeclarations ']'
        '{' declarations '}'	          { endProto();   
                                            $$=pt_node(PT_PROTO,$2,$5,$8);
                                            myfree($2);}
	;

externproto: EXTERNPROTO IDENTIFIER	             { beginProto($2); }
             '[' externInterfaceDeclarations ']'     { expect(MFSTRING); }
             fieldValue				     { endProto();
                                                       $$=pt_node(PT_EXTPROTO,$2,$5,$8);
                                                       myfree($2); }
	    ;


interfaceDeclarations: /* Empty is OK */      { $$=pt_node(PT_INTERFACE,NULL); } 
	|	interfaceDeclarations interfaceDeclaration
                         { $1->addChild($2); 
                           $$=$1; }
        ;

interfaceDeclaration: EVENTIN IDENTIFIER IDENTIFIER  { addEventIn($2, $3);
                                                       tmppt=pt_node(PT_ID,$2);
                                                       $$=pt_node(PT_EVENTIN,$3,tmppt);
                                                        myfree($2); myfree($3); }
   | /* STED */
     EVENTIN IDENTIFIER IDENTIFIER  IS routeEvent { addEventIn($2, $3);
                                                    tmppt=pt_node(PT_ID,$2);
                                                    $$=pt_node(PT_EVENTIN_IS,$3,tmppt,$5);
                                                    check_subtype($5->classname,
                                                                  $2);
                                                    myfree($2); myfree($3); }
   |   EVENTOUT IDENTIFIER IDENTIFIER		{ addEventOut($2, $3);
                                                  tmppt=pt_node(PT_ID,$2);
                                                  $$=pt_node(PT_EVENTOUT,$3,tmppt);
 						  myfree($2); myfree($3); }
   | /* STED */
     EVENTOUT IDENTIFIER IDENTIFIER IS routeEvent { addEventOut($2, $3);
                                                    tmppt=pt_node(PT_ID,$2);
                                                    $$=pt_node(PT_EVENTOUT_IS,$3,tmppt,$5);
                                                    check_subtype($5->classname,
                                                                  $2);
 						    myfree($2); myfree($3); }
   |   FIELD IDENTIFIER IDENTIFIER		{ int type = fieldType($2);
                                                  expect(type); }

      	fieldValue        			{ tmppt=pt_node(PT_ID,$2);
                                                  check_subtype($5->classname,
                                                                $2);
                                                  $$=pt_node(PT_FIELD,$3,tmppt,$5);
                                                  addField($2, $3, $5);
                                                  myfree($2); myfree($3); }
    |	EXPOSEDFIELD IDENTIFIER	IDENTIFIER	{ int type = fieldType($2);
    						  expect(type); }
                     
    	fieldValue                              { tmppt=pt_node(PT_ID,$2);
                                                  check_subtype($5->classname,
                                                                $2);
                                                  $$=pt_node(PT_EXPFIELD,$3,tmppt,$5);
                                                  addExposedField($2, $3, $5);
                                                  myfree($2); myfree($3); }
    ;

externInterfaceDeclarations: /* Empty is OK */   { $$=pt_node(PT_EXTINTERFACE,NULL); } 
     |  externInterfaceDeclarations externInterfaceDeclaration
                       { $1->addChild($2); 
                         $$=$1; }
     ;

externInterfaceDeclaration: EVENTIN IDENTIFIER IDENTIFIER   { addEventIn($2, $3);
                                                              tmppt=pt_node(PT_ID,$2);
                                                              $$=pt_node(PT_EVENTIN,$3,tmppt);
        						      myfree($2); myfree($3); }
    |   EVENTOUT IDENTIFIER IDENTIFIER		{ addEventOut($2, $3);
                                                  tmppt=pt_node(PT_ID,$2);
                                                  $$=pt_node(PT_EVENTOUT,$3,tmppt);
        			       		  myfree($2); myfree($3); }
    |	FIELD IDENTIFIER IDENTIFIER	        { addField($2, $3, NULL);
                                                  tmppt=pt_node(PT_ID,$2);
                                                  $$=pt_node(PT_FIELD,$3,tmppt);
						  myfree($2); myfree($3); }
    |	EXPOSEDFIELD IDENTIFIER IDENTIFIER	{ addExposedField($2, $3, NULL);
                                                  tmppt=pt_node(PT_ID,$2);
                                                  $$=pt_node(PT_EXPFIELD,$3,tmppt);
						  myfree($2); myfree($3); }
    ;

/*<< STED */
routeDeclaration: ROUTE routeEvent TO routeEvent     
         { $$=pt_node(PT_ROUTE,NULL,$2,$4);
           check_subtype($2->classname,$4->classname);
         }
   | UNROUTE routeEvent TO routeEvent     
         { $$=pt_node(PT_UNROUTE,NULL,$2,$4);
           check_subtype($2->classname,$4->classname);
         }
          ; 

routeEvent: IDENTIFIER   { VrmlNodeType::NameTypeRec* nt=getFieldorEvent($1);
                           $$=pt_node(PT_R_EVENT_A,$1);
                           if (nt)
                             $$->classname=nt->typestring;
                           else
                            { $$->classname="UNKNOWN";
                              yyerror($1);  
                              cerr << " is not a local field or event !\n";
                            }
                           myfree($1); }
     |      IDENTIFIER '.' IDENTIFIER 
          { int ftype;
            char* node_classname;
            VrmlNodeType *t;
            tmppt=pt_node(PT_R_EVENT_A,$3);
            if (ftype=hasFieldorEvent($1))
              { VrmlNodeType::NameTypeRec* nt=getFieldorEvent($1);                                           $$=pt_node(ftype,$1,tmppt);
                node_classname=nt->typestring;
	      }
            else
              { DefNode* dp=finddef($1);
                $$=pt_node(PT_R_EVENT_C,$1,tmppt);
                if (dp)
                   node_classname=dp->type;
                 else
                   node_classname="UNKNOWN";
              }
          t = VrmlNodeType::find(node_classname);
          if (t)
            { VrmlNodeType::NameTypeRec* nt=getFieldorEvent(t,$3);
              if (nt)
                 $$->classname=nt->typestring;
                else
                 { $$->classname="UNKNOWN";
                   yyerror($1);
                   cerr << " has no field or event " 
                        << $3 << " !\n";
                 }
            }
           else $$->classname="UNKNOWN";
          myfree($1); myfree($3); }
     ;

/* original rule was 
// routeDeclaration:
//		ROUTE IDENTIFIER '.' IDENTIFIER TO IDENTIFIER '.' IDENTIFIER
//        		{ myfree($2); myfree($4); myfree($6); myfree($8); }
//  ; 
*/

/* STED >>*/

node:	IDENTIFIER  				{ enterNode($1); }
      	'{' nodeGuts '}'	        	{ exitNode();
                                                  $$=pt_node(PT_NODE,$1,$4);
                                                  $$->classname=strdup($1);
                                                  myfree($1);}
    ;

nodeGuts: /* Empty is OK */ { $$=pt_node(PT_GUTS,NULL); } 
    | 	nodeGuts nodeGut    { $1->addChild($2); 
                              $$=$1; }
    ;

nodeGut: IDENTIFIER		{ enterField($1); }
         fieldValue      	{ VrmlNodeType::NameTypeRec* nt;
                                  nt=getFieldorEvent($1);
                                  if (nt)
                                     { check_subtype($3->classname,
                                                     nt->typestring);
                                     }
                                  exitField(); 
                                 
                                  $$=pt_node(PT_IDENTIFIER,$1,$3);
                                  myfree($1); }
    |	routeDeclaration
    |	protoDeclaration

    	/* The following are only valid for Script nodes: */
    |   EVENTIN IDENTIFIER IDENTIFIER		{ inScript();
                                                  tmppt=pt_node(PT_ID,$2); 
                                                  $$=pt_node(PT_EVENTIN,$3,tmppt);
                                                  myfree($2); myfree($3); }
    |   EVENTOUT IDENTIFIER IDENTIFIER		{ inScript();
                                                  tmppt=pt_node(PT_ID,$2); 
                                                  $$=pt_node(PT_EVENTOUT,$3,tmppt);
                                                  myfree($2); myfree($3); }
    |   FIELD IDENTIFIER IDENTIFIER	        { inScript(); int type = fieldType($2);
    							      expect(type); }
        fieldValue                              { tmppt=pt_node(PT_ID,$2);
                                                  check_subtype($5->classname,
                                                                $2); 
                                                  $$=pt_node(PT_FIELD,$3,tmppt,$5);
                                                  myfree($2); myfree($3); }
    |	/* STED was: EVENTIN IDENTIFIER IDENTIFIER IS IDENTIFIER */
        EVENTIN IDENTIFIER IDENTIFIER IS routeEvent
    			{ inScript(); 
                          tmppt=pt_node(PT_ID,$2); 
                          check_subtype($5->classname,
                                        $2);
                          $$=pt_node(PT_EVENTIN_IS,$3,tmppt,$5);
                          myfree($2); myfree($3); }
    |	/* STED was EVENTOUT IDENTIFIER IDENTIFIER IS IDENTIFIER */
        EVENTOUT IDENTIFIER IDENTIFIER IS routeEvent
    			{ inScript();
                          tmppt=pt_node(PT_ID,$2);
                          check_subtype($5->classname,
                                        $2); 
                          $$=pt_node(PT_EVENTOUT_IS,$3,tmppt,$5);
                          myfree($2); myfree($3); }
   ;

fieldValue: SFBOOL { $$=$1; $$->classname="SFBool"; }
    |   SFCOLOR    { $$=$1; $$->classname="SFColor"; }
    |	MFCOLOR    { $$=$1; $$->classname="MFColor"; }
    |   SFFLOAT    { $$=$1; $$->classname="SFFloat"; }
    |	MFFLOAT    { $$=$1; $$->classname="MFFloat"; }
    |	SFIMAGE    { $$=$1; $$->classname="SFImage"; }
    |	SFINT32    { $$=$1; $$->classname="SFInt32"; }
    |	MFINT32    { $$=$1; $$->classname="MFInt32"; }
    |	SFROTATION { $$=$1; $$->classname="SFRotation"; }
    |	MFROTATION { $$=$1; $$->classname="MFRotation"; }
    |	SFSTRING   { $$=$1; $$->classname="SFString"; }
    |	MFSTRING   { $$=$1; $$->classname="MFString"; }
    |	SFTIME     { $$=$1; $$->classname="SFTime"; }
    |	SFVEC2F    { $$=$1; $$->classname="SFVec2f"; }
    |	MFVEC2F    { $$=$1; $$->classname="MFVec2f"; }
    |	SFVEC3F    { $$=$1; $$->classname="SFVec3f"; }
    |	MFVEC3F    { $$=$1; $$->classname="MFVec3f"; }
    
  /* STED propagate types
   */ 
    |	SFNODE nodeDeclaration { $$=pt_node(PT_SFNODE,NULL,$2);
                                 $$->classname=$2->classname; }
    |	SFNODE SFN_NULL        { $$=pt_node(PT_SFN_NULL,NULL);
                                 $$->classname="NULL";
                               }
    |	MFNODE mfnodeValue     { $$=pt_node(PT_MFNODE,NULL,$2); 
                                 $$->classname="MFNode"; }
    |	/* STED was IS IDENTIFIER */
        /* note: this is not yet converted to VRML 2.0,
           only for the case that routeEvent is a single IDENTIFIER */
       IS routeEvent	       { $$=pt_node(PT_IS,NULL,$2); 
                                 $$->classname=$2->classname; }
    ;

mfnodeValue: '[' nodes ']'   { $$=pt_node(PT_MFNODE2,NULL,$2); }
     |  nodeDeclaration      { $$=$1; }
     ;

nodes:	/* Empty is OK */      { $$=pt_node(PT_DECS,NULL); } 
     |  nodes nodeDeclaration  { $1->addChild($2); 
                                 $$=$1; }
     ;


/*<< STED */ 
 classDeclaration: class        { $$=$1; }
        |          externclass  { $$=$1; }
        ;

 class: CLASS IDENTIFIER                         { beginClass($2); }	  
        '[' interfaceDeclarations ']' extends
        '{' declarations '}'                     { endClass();
                                                   $$=pt_node(PT_CLASS,$2,$5,$7,$9);
                                                   myfree($2); }	  
        ;

 externclass: EXTERNCLASS IDENTIFIER		    { beginClass($2); }		
	      '[' externInterfaceDeclarations ']' 
              extends           		    { expect(MFSTRING); }
            fieldValue				    { endClass();
                                                      $$=pt_node(PT_EXTCLASS,$2,$5,$7,$9);
                                                      myfree($2); }
	;

 extends: EXTENDS superclasses { $$=$2; }

 superclasses: IDENTIFIER                  { addSuperclass($1);
                                             tmppt=pt_node(PT_ID,$1);
                                             $$=pt_node(PT_SUPERS,NULL,tmppt);
                                             myfree($1); }
       |       IDENTIFIER superclasses     { addSuperclass($1);
                                             tmppt=pt_node(PT_ID,$1);
                                             $2->addChild_front(tmppt);
                                             $$=$2;
                                             myfree($1); }
       ;
 
 
/* STED >>*/


%%

extern char* yytext;

extern int column;

void err_indent(int d) { while(d--) { cerr << " "; }}

extern char* linebuffer;

void 
yyerror(const char *msg)
{     cerr << "\n" << linebuffer << "\n";
      err_indent(column);
      cerr << "^\n";
      cerr << "Error near line " << currentLineNumber << ": " << msg;
      expect(0);
}

int hasFieldorEvent(const char* name)
{ VrmlNodeType *t;
  int ft;
  if (currentProtoStack.empty()) return 0;
  t = currentProtoStack.top();
  
  if (ft=t->hasField(name))
    { if (ft==SFNODE) return PT_ROUTE_EVENT_F_SF;
      if (ft==MFNODE) return PT_ROUTE_EVENT_F_MF;
    }
  if (ft=t->hasExposedField(name))
    { if (ft==SFNODE) return PT_ROUTE_EVENT_EX_SF;
      if (ft==MFNODE) return PT_ROUTE_EVENT_EX_MF;
    }
  if (ft=t->hasEventIn(name))
    { if (ft==SFNODE) return PT_ROUTE_EVENT_EI_SF;
      if (ft==MFNODE) return PT_ROUTE_EVENT_EI_MF;
    }
  if (ft=t->hasEventOut(name))
   { if (ft==SFNODE) return PT_ROUTE_EVENT_EO_SF;
     if (ft==SFNODE) return PT_ROUTE_EVENT_EO_MF;
   }
  return 0;
}

VrmlNodeType::NameTypeRec* getFieldorEvent(const char* name)
{ VrmlNodeType *t;
  FieldRec* fr;
  if (!currentField.empty())
    { VrmlNodeType::NameTypeRec* res;
      fr=currentField.top();
      res=getFieldorEvent((VrmlNodeType*)fr->nodeType,name);
      if (res) return res;
    }
  if (currentProtoStack.empty()) return NULL;
     t = currentProtoStack.top();
  return getFieldorEvent(t,name);
}

VrmlNodeType::NameTypeRec* getFieldorEvent(VrmlNodeType *t, const char* name)
{ VrmlNodeType::NameTypeRec *res;
  list<VrmlNodeType*>::iterator i;
  if (res=t->get(t->getEventIns(),name)) return res;
  if (res=t->get(t->getEventOuts(),name)) return res;
  if (res=t->get(t->getFields(),name)) return res;
  if (t->isClass())
    { list<VrmlNodeType*> supers=t->getSuperClasses();
      for(i=supers.begin();i!=supers.end();i++)
         if (res=getFieldorEvent((*i),name)) return res;
    }
  return(NULL);
}

static void
beginProto(const char *protoName)
{
	// Any protos in the implementation are in a local namespace:
    VrmlNodeType::pushNameSpace();
    deflist.push_front((DefNode*) NULL);

    VrmlNodeType *t = new VrmlNodeType(protoName);
    currentProtoStack.push(t);
}

static void
beginClass(const char *protoName)
{
    // Any protos in the implementation are in a local namespace:
    VrmlNodeType::pushNameSpace();
    deflist.push_front((DefNode*) NULL);

    VrmlNodeType *t = new VrmlNodeType(protoName,1);
    currentProtoStack.push(t);
}

static void
endProto()
{
    // Make any protos defined in implementation unavailable:
    VrmlNodeType::popNameSpace();

    // Remove everything up to and including the next NULL marker:
    list<DefNode*>::iterator i;
    for (i = deflist.begin(); i != deflist.end();) {
    	DefNode *nodeType = *i;
        ++i;
        deflist.pop_front();

        if (nodeType ==  NULL) {
            break;
        }
        else { delete nodeType;
        }
      }

    // Add this proto definition:
	if (currentProtoStack.empty()) {
    	cerr << "Error: Empty PROTO stack!\n";
    }
    else {
    	VrmlNodeType *t = currentProtoStack.top();
    	currentProtoStack.pop();
        VrmlNodeType::addToNameSpace(t);
    }
}

static void
endClass()
{ endProto();
}

int
addField(const char *type, const char *name, ParseTreeNode* pt)
{
	return add(&VrmlNodeType::addField, type, name, pt);
}

int
addEventIn(const char *type, const char *name)
{
	return add(&VrmlNodeType::addEventIn, type, name);
}
int
addEventOut(const char *type, const char *name)
{
	return add(&VrmlNodeType::addEventOut, type, name);
}
int
addExposedField(const char *type, const char *name, ParseTreeNode* pt)
{
	return add(&VrmlNodeType::addExposedField, type, name, pt);
}

void addSuperclass(const char* super)
 {  VrmlNodeType *t = currentProtoStack.top();
    t->addSuperclass(super);
 }

int
add(void (VrmlNodeType::*func)(const char *, int, const char*, ParseTreeNode*), 
	const char *typeString, const char *name,
        ParseTreeNode* pt)
{
    int type = fieldType(typeString);

    if (type == 0) {
    	cerr << "Error: invalid field type: " << type << "\n";
    }

    // Need to add support for Script nodes:
	// if (inScript) ... ???

	if (currentProtoStack.empty()) {
    	cerr << "Error: declaration outside of prototype\n";
        return 0;
    }
   VrmlNodeType *t = currentProtoStack.top();
    (t->*func)(name, type, typeString, pt);
    
    return type;
}

int
add(void (VrmlNodeType::*func)(const char *, int, const char*), 
	const char *typeString, const char *name)
{
    int type = fieldType(typeString);

    if (type == 0) {
    	cerr << "Error: invalid field type: " << type << "\n";
    }

    // Need to add support for Script nodes:
	// if (inScript) ... ???

	if (currentProtoStack.empty()) {
    	cerr << "Error: declaration outside of prototype\n";
        return 0;
    }
    VrmlNodeType *t = currentProtoStack.top();
    (t->*func)(name, type, typeString);
    
    return type;
}


int
fieldType(const char *type)
{
    if (strcmp(type, "SFBool") == 0) return SFBOOL;
    if (strcmp(type, "SFColor") == 0) return SFCOLOR;
    if (strcmp(type, "SFFloat") == 0) return SFFLOAT;
    if (strcmp(type, "SFImage") == 0) return SFIMAGE;
    if (strcmp(type, "SFInt32") == 0) return SFINT32;
    if (strcmp(type, "SFNode") == 0) return SFNODE;
    if (strcmp(type, "SFRotation") == 0) return SFROTATION;
    if (strcmp(type, "SFString") == 0) return SFSTRING;
    if (strcmp(type, "SFTime") == 0) return SFTIME;
    if (strcmp(type, "SFVec2f") == 0) return SFVEC2F;
    if (strcmp(type, "SFVec3f") == 0) return SFVEC3F;
    if (strcmp(type, "MFColor") == 0) return MFCOLOR;
    if (strcmp(type, "MFFloat") == 0) return MFFLOAT;
    if (strcmp(type, "MFInt32") == 0) return MFINT32;
    if (strcmp(type, "MFNode") == 0) return MFNODE;
    if (strcmp(type, "MFRotation") == 0) return MFROTATION;
    if (strcmp(type, "MFString") == 0) return MFSTRING;
    if (strcmp(type, "MFVec2f") == 0) return MFVEC2F;
    if (strcmp(type, "MFVec3f") == 0) return MFVEC3F;
    
    VrmlNodeType *t = VrmlNodeType::find(type);
    if /* ((t!=NULL) && (t->isClass())) */
       /* we allow also PROTOs as Type Restrictions*/
      (t!=NULL)
      { return SFNODE;
      }

    cerr << "Illegal field type: " << type << "\n";

    return 0;
}

void
enterNode(const char *nodeType)
{
    VrmlNodeType *t = VrmlNodeType::find(nodeType);
    if (t == NULL) {
    	char tmp[1000];
        VrmlNodeType::print_types();
        sprintf(tmp, "Unknown node type '%s'", nodeType);
    	yyerror(tmp);
    }
    FieldRec *fr = new FieldRec;
    fr->nodeType = t;
    fr->fieldName = NULL;
    currentField.push(fr);
}

void
exitNode()
{
	FieldRec *fr = currentField.top();
    assert(fr != NULL);
	currentField.pop();

    delete fr;
}

void
inScript()
{
	FieldRec *fr = currentField.top();
    if (fr->nodeType == NULL ||
    	strcmp(fr->nodeType->getName(), "Script") != 0) {
        yyerror("interface declaration outside of Script or prototype");
    }
}

void
enterField(const char *fieldName)
{
    FieldRec *fr = currentField.top();
    assert(fr != NULL);

    fr->fieldName = fieldName;
    if (fr->nodeType != NULL) {
    	// enterField is called when parsing eventIn and eventOut IS
        // declarations, in which case we don't need to do anything special--
        // the IS IDENTIFIER will be returned from the lexer normally.
    	if (fr->nodeType->hasEventIn(fieldName) ||
            fr->nodeType->hasEventOut(fieldName))
           return;
    
        int type = fr->nodeType->hasField(fieldName);
        if (type != 0) {
        	// Let the lexer know what field type to expect:
        	expect(type);
        }
        else {
        	cerr << "Error: Node's of type " << fr->nodeType->getName() <<
            	" do not have fields/eventIn/eventOut named " <<
                fieldName << "\n";
            // expect(ANY_FIELD);
        }
    }
    // else expect(ANY_FIELD);
}

void
exitField()
{
	FieldRec *fr = currentField.top();
    assert(fr != NULL);

	fr->fieldName = NULL;
}

void
expect(int type)
{
	expectToken = type;
}

DefNode* finddef(const char* name)
{  // Look through the deflistL
  list<DefNode*>::iterator i;
  for (i = deflist.begin(); i != deflist.end(); i++) {
    	DefNode *nt = *i;
    	if ((nt !=  NULL) && 
            (strcmp(nt->name,name)) == 0) {
            return nt;
        }
    }
    return  NULL;
}

int subtype(const char* t1, const char* t2)
{ VrmlNodeType *tn1, *tn2; 
  
  if (strcmp(t1,t2)==0) return(1);
  if (strcmp(t1,"NULL")==0) return(1);
  if ((strcmp(t2,"SFNode")==0) || (strcmp(t2,"MFNode")==0))
     return(1);
  if ((strcmp(t1,"SFNode")==0) || (strcmp(t1,"MFNode")==0))
     return(0);
  tn1= VrmlNodeType::find(t1); 
  tn2= VrmlNodeType::find(t2);
  if ((!tn1) || (!tn2)) return(0);
  
  if (tn1->isClass())
    { list<VrmlNodeType*> supers=tn1->getSuperClasses();
      list<VrmlNodeType*>::const_iterator j;
      for (j = supers.begin(); j != supers.end(); j++) 
        if (subtype((*j)->getName(),t2)) return(1);
    }
  return(0);
}

int check_subtype(const char* t1, const char* t2)
{ int ant=2;
#ifdef TCHECKDEBUG 
 cerr << currentLineNumber << " subtype: " << t1 << " " << t2 << " \n";
#endif
  if ((strcmp(t1,"UNKNOWN")==0) 
     || (strcmp(t2,"UNKNOWN")==0)) ant=1;
  else
  if ((t2[0]=='M') && (t2[1]=='F'))
    { if (strcmp(t1+2,t2+2)==0) ant=1;
    }
  else
  if ((t1[0]=='S') && (t1[1]=='F')
      && (t2[0]=='S') && (t2[1]=='F'))
	{ if (strcmp(t1+2,t2+2)==0) ant=1;
          else ant=0;
        }
  else ant=subtype(t1,t2);
  if (!ant)
     { yyerror(t1); 
       cerr << " is not a subtype of " 
            << t2 << " !\n";
     }
  return(ant);
}










